/**************************************************
Shared Sketch by Active Bases --- Testing Phase
   - Invariant to rotation and shift -


Input:  template.
Output: Rotation, shift, and score for each patch.

***************************************************/

# include <stdio.h>
# include <stdlib.h>
# include "mex.h"        /* the algorithm is connect to matlab */
# include "math.h"
# define PI 3.1415926
# define ABS(x) ((x)>0? (x):(-(x)))
# define MAX(x, y) ((x)>(y)? (x):(y))
# define MIN(x, y) ((x)<(y)? (x):(y))

double *double_vector(int n)
{
    double *v;    
    v = (double*) mxCalloc (n, sizeof(double));
    return v; 
}

int *int_vector(int n)
{
    int *v; 
    v = (int*) mxCalloc (n, sizeof(int));
    return v; 
}

double **double_matrix(int m, int n)
{
    double **mat; 
    int i; 
    mat = (double**) mxCalloc(m, sizeof(double*)); 
    for (i=0; i<m; i++)
        mat[i] = double_vector(n);
    return mat; 
}

int **int_matrix(int m, int n)
{
    int **mat; 
    int i; 
    mat = (int**) mxCalloc(m, sizeof(int*)); 
    for (i=0; i<m; i++)
        mat[i] = int_vector(n); 
    return mat;
}

/* getting the index of matlab image, (x,y) location, (sx, sy) sizes */
int px(int x, int y, int bx, int by)    
{            
   return (x + (y-1)*bx - 1); 
}

 int n;                      /* number of images */
 int N;                      /* number of orientations */
 double **fI;                /* filtered images */
 double **C;                 /* inhibition coefficients */
 int h;                      /* halfsizes of filters */
 int sx, sy;                 /* sizes of image */ 
 double T;                   /* threshold of selected bases */
 double *Mi, *Mx, *My, *Mm, *Mm1;  /* storing selected bases */
 double **allsymbol;         /* symbol of filters */
 double *sym, **Asym;        /* symbol of selected Gabors */
 int L, ore;                 /* allowed ranges of shifting in location and orientation */
 int sub;                    /* subsampling */
 double *tss;                /* log-likelihood score for each testing image */
 double Upperbound;          /* satuation level for response */
 int Totalsketch;            /* total number of bases */
 double *responseVec;	     /* response vector for each testing image */
 int **sinsh, **cossh;       /* store the shift to avoid repeated sin and cosin computation */
 int M;
 double *lam, *e, *lz;
 double *shiftX, *shiftY, *rotation;
 
/* store the shift values so that we do not need to repeat the sin and cos computation */ 
void storeshift()
{
    int ind, l;
    double theta; 
    
    sinsh = int_matrix(N, L+L+1); 
    cossh = int_matrix(N, L+L+1); 
    for (ind=0; ind<N; ind++)        
    {
        theta = PI*ind/N; 
        for (l=-L; l<=L; l++)
         {
            sinsh[ind][l+L] = floor(l*sub*sin(theta)+.5); 
            cossh[ind][l+L] = floor(l*sub*cos(theta)+.5); 
        }   
    }
}
    
/* for Gabor(x, y, orientation = ind), find local maximum in image i */
double shiftmax(int i, int ind, int x, int y, int *rm) 
{
   double m;
   int x1, y1, l, ml, mx, my, de, d, d1, here, mo, md; 
 
   m = 0.;
   for (l=-L; l<=L; l++)   
     {
        x1 = x + cossh[ind][l+L]; 
        y1 = y + sinsh[ind][l+L];   /* shifting in normal direction */
        if ((x1>=1)&&(x1<=sx)&&(y1>=1)&&(y1<=sy))
        {
        here = px(x1, y1, sx, sy);
        for (de=-ore; de<=ore; de++)
            {
              d = de+ind;
              d1 = d;
              if (d<0)
                 d1 = d+N;
              if (d>=N)
                  d1 = d-N;   /* shifting in orientation */
              if (fI[d1*n+i][here]>m)
                {
                   m = fI[d1*n+i][here];   /* local maximization */
                   ml = l; mx = x1; my = y1; mo = d1; md = de; 
                }
             }
        }
     }
   rm[0] = ml; rm[1] = mx; rm[2] = my; rm[3] = mo; rm[4] = md; 
   return(MIN(Upperbound, m));
}

/* the Gabor(mx, my, orientation = mi) inhibits overlapping Gabors for image i */
void inhibit(int i, int mi, int mx, int my) 
{
   int x, y, ind;
   double *f, *fc; 
   
   for (ind=0; ind<N; ind++)   
   {
       f = fI[ind*n+i];   
       fc = C[mi+ind*N];   /* inhibition coefficients, zero = inhibit */
       for (x=MAX(1, mx-2*h); x<=MIN(sx, mx+2*h); x++)
         for (y=MAX(1, my-2*h); y<=MIN(sy, my+2*h); y++)
         {
          f[px(x, y, sx, sy)] *= fc[px(x-mx+2*h+1, y-my+2*h+1, 4*h+1, 4*h+1)];
         }
   }
}


/* draw the symbol of Gabor(x0, y0, orientation = ind) with intensity w */
void draw(double *sym, int x0, int y0, int ind, double w)
{
  int x, y; 
  double a;
          
  for (x=x0-h; x<=x0+h; x++)
     for (y=y0-h; y<=y0+h; y++)
       {
         a = allsymbol[ind][px(x-x0+h+1, y-y0+h+1, 2*h+1, 2*h+1)]*w; 
         if (sym[px(x, y, sx, sy)]<a)
             sym[px(x, y, sx, sy)] = a;
       }
}

/*
	(cx,cy): pivot of rotation
	(x,y,iOri): point to be transformed
	(deltaX,deltaY,deltaOri): the transform
	*xyo[3]: point after transformation
*/
void geometricTransform(int cx,int cy,int x,int y,int iOri,int deltaX,int deltaY,int deltaOri,int *xyo)
{
	double alpha,radius;
	// First, rotate
	xyo[2] = iOri+deltaOri;
	while(xyo[2]<0)
		xyo[2] += N;
	if(xyo[2]>=N)
		xyo[2] = xyo[2] % N;
	alpha = atan2(y-cy,x-cx) + deltaOri*PI/N;
	radius = sqrt((y-cy)*(y-cy)+(x-cx)*(x-cx));
	// Then, shift the center
	cx += deltaX; cy += deltaY;
	xyo[0] = floor(0.5 + cx + radius*cos(alpha));
	xyo[1] = floor(0.5 + cy + radius*sin(alpha));
}

/* the shared sketch algorithm */
void Cmp()
{
   int i, t, rm[5], xyo[3], k; 
   double mm, *tmp;
   int j, deltaOri, deltaX, deltaY; j=0;
   tmp = double_vector(n*(4*2)*3*3);
	 for (i=0; i<n; i++) // for each image
	 {
		k = 0;
		for(deltaOri=0;deltaOri<N*2;deltaOri+=4)
			for(deltaX=-1;deltaX<=1;deltaX++)
				for(deltaY=-1;deltaY<=1;deltaY++)
				{
					for(t=0;t<Totalsketch;t++)
					{
						geometricTransform(floor(sx/2),floor(sy/2),Mx[t],My[t],Mi[t],deltaX,deltaY,deltaOri,xyo);
						mm = shiftmax(i, xyo[2], xyo[0], xyo[1], rm);
						tmp[k*n+i] += Mm[t]*mm - Mm1[t];
					}
					k++;
				}
	 }
	 for(i=0;i<n;i++)
	 {
		 k = 0;
		 tss[i] = -1000;
		 for(deltaOri=0;deltaOri<N*2;deltaOri+=4)
			for(deltaX=-1;deltaX<=1;deltaX++)
				for(deltaY=-1;deltaY<=1;deltaY++)
				{
					if(tss[i]<tmp[k*n+i])
					{
						tss[i] = tmp[k*n+i];
						shiftX[i] = deltaX;
						shiftY[i] = deltaY;
						rotation[i] = deltaOri;
					}
					k++;
				}
		//if (mm[i]>0.)
		//{
           // draw(Asym[i], rm[1], rm[2], rm[3], sqrt(mm)); 
		   // inhibit(i, rm[3], rm[1], rm[2]); // In CFind, inhibition is not used !
		//}
	 }
}


/* mex function is used to pass on the pointers and scalars from matlab, 
   so that heavy computation can be done by C, which puts the results into 
   some of the pointers. After that, matlab can then use these results. 
   
   So matlab is very much like a managing platform for organizing the 
   experiments, and mex C is like a work enginee for fast computation. */

void mexFunction(int nlhs, mxArray *plhs[], 
                 int nrhs, const mxArray *prhs[])                
{
 int ind, i, j, c; 
 mxArray *f;  
 
 c = 0; /* counter for input pointers and scalars */
 n = floor(mxGetScalar(prhs[c++])+.5);  /* number of images */
 N = floor(mxGetScalar(prhs[c++])+.5);  /* number of orientations */
 fI = mxCalloc(n*N, sizeof(double*));   /* fitered images */
 for (i=0; i<n; i++)
   {
     for (ind=0; ind<N; ind++)
      {  
       f = mxGetCell(prhs[c], ind*n+i); 
       fI[ind*n+i] = mxGetPr(f);    /* get pointers to filtered images */       
      }
    }
 c++; 

 
 C = mxCalloc(N*N, sizeof(double*));    /* C: correlation/inhibition between filters */
 for (ind=0; ind<N; ind++)
     {  
       for (j=0; j<N; j++)
        {
         f = mxGetCell(prhs[c], j*N+ind); 
         C[j*N+ind] = mxGetPr(f);         /* get correlation/inhibition */
        }   
     }
 c++; 
 
 h = floor(mxGetScalar(prhs[c++])+.5);    /* half size of filters */
 
 sx = floor(mxGetScalar(prhs[c++])+.5);
 sy = floor(mxGetScalar(prhs[c++])+.5);   /* size of images */
 T = mxGetScalar(prhs[c++]);              /* threshold */  
     
 Mi = mxGetPr(prhs[c++]);   /* orientation of selected Gabor */               
 Mx = mxGetPr(prhs[c++]);         
 My = mxGetPr(prhs[c++]);   /* (x, y) position of selected Gabor */
 Mm = mxGetPr(prhs[c++]);   /* lambda */    
 Mm1 = mxGetPr(prhs[c++]);  /* logZ */ 

 allsymbol = mxCalloc(N, sizeof(double*));    
 for (ind=0; ind<N; ind++)
     {  
       f = mxGetCell(prhs[c], ind); 
       allsymbol[ind] = mxGetPr(f);  /* symbols of filters */         
     }
 c++; 
     
 Asym = mxCalloc(n, sizeof(double*));   /* symbols of active Gabors for an individual image */
 for (i=0; i<n; i++)
    {
        f = mxGetCell(prhs[c], i);
        Asym[i] = mxGetPr(f);  
     }
 c++; 
 
 L = floor(mxGetScalar(prhs[c++])+.5);     /* range of shifting along normal direction of Gabor */
 ore = floor(mxGetScalar(prhs[c++])+.5);   /* range of shifting in orientation */
 sub = floor(mxGetScalar(prhs[c++])+.5);   /* sub-sampling of Gabor filters to improve speed */
 Totalsketch = floor(mxGetScalar(prhs[c++]));  /* total number of bases */
 Upperbound = mxGetScalar(prhs[c++]);             /* satuation level */
 tss = mxGetPr(prhs[c++]);                        /* map of log-likelihood score */

 responseVec = mxGetPr(prhs[c++]);
 
 M = floor(mxGetScalar(prhs[c++])+.5);  
 lam = mxGetPr(prhs[c++]);   
 e = mxGetPr(prhs[c++]);   
 lz = mxGetPr(prhs[c++]);
 shiftX = mxGetPr(prhs[c++]);
 shiftY = mxGetPr(prhs[c++]);
 rotation = mxGetPr(prhs[c++]);
 
 storeshift();
 Cmp();
}

                    