clear all

mex geoTransform.c
mex CTesting.c
posFolder = 'horse';
% CONSTANTS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
scale = .7;  % scale of Gabors, length = 17 pixels
norient = 15;  % number of orientations
T = 0.;        % threshold for responses
sx = 100; sy = 120;
% get gabor responses (S1 maps)
[allfiltered allfiltered_r allsymbol allfilter images] = getGaborResponse(posFolder,T,scale,norient,sx,sy);
nimage = size(allfiltered,1);
%reflect the S1 maps
allfiltered_r = cell(nimage, norient); % reflected S1 maps
for i = 1:nimage
    for j = 1:norient
        srcImage = allfiltered{i,j};
        destImage = zeros(size(srcImage));
        for col = 1:size(srcImage,2)
            destImage(:,size(srcImage,2)-col+1) = srcImage(:,col);
        end
        rf = norient+2-j;
        if j == 1
            rf = 1;
        end
        allfiltered_r{i, rf} = destImage;
    end
end

nimage = size(allfiltered,1);


w = rand(1,nimage); % probability that an image belongs to the canonical pose.
                    % randomly initialized
save('init.mat','w');
step = 1;
while step<=3
    tic
    % concatenate data weights and S1 maps of two supperposed states
    w_con = [w 1-w];
    allfiltered_con = cell(2*nimage,norient);
    for i = 1:nimage
        for j = 1:norient
            allfiltered_con{i,j} = allfiltered{i,j}+0;
            % reflect S1 map
            srcImage = allfiltered{i,j};
            destImage = zeros(size(srcImage));
            for col = 1:size(srcImage,2)
                destImage(:,size(srcImage,2)-col+1) = srcImage(:,col);
            end
            rf = norient+2-j;
            if j == 1
                rf = 1;
            end
            %allfiltered_con{i+nimage, norient-j+1} = destImage;
            allfiltered_con{i+nimage, rf} = destImage;
        end
    end
    [learnedModel, Asym] = ABlearn(w_con,allfiltered_con,allsymbol,allfilter,T); % learn only one template
    towrite = learnedModel.sym;
    towrite = uint8(255 * (towrite-min(towrite(:)))/(max(towrite(:))-min(towrite(:))));
    imwrite(towrite, ['template' num2str(step) '.bmp']);
      outepsgif(-learnedModel.sym, ['ex_flip_' posFolder '_template' num2str(step)]); 

  
    disp(['training time (M-step): ' num2str(toc) ' seconds']);
    tic
    tss = learnedModel.tss;
    tss_r = tss(nimage+(1:nimage));
    tss = tss(1:nimage);
%     tss_r = ABtest_one(learnedModel,allsymbol,allfilter,allfiltered_r);
    disp(['testing time (E-step): ' num2str(toc) ' seconds']);
    dif = 100;
    old_w = w;
    a = sum(w(:))/nimage;
    w = a*exp(tss)./(a*exp(tss)+(1-a)*exp(tss_r));

    if step > 1
        dif = sum((old_w-w).^2)/nimage;
    end
    if step>1 && dif < 1e-6
        break
    end

    step = step + 1;
end

save('model.mat','learnedModel');

ind = find(w>=0.5);
canonicalPose = {images{ind}};
canonicalAsym = {Asym{ind}};
im = displayImagesAndSketches(canonicalPose, canonicalAsym, 5, sx, sy);
outepsgif(im, ['ex_flip_' posFolder '_canonical']); 

im = displayOverlay(canonicalPose, canonicalAsym, 8, sx, sy);
imwrite(im,'canonical_overlay.bmp');
outepsgif(im, ['ex_flip_' posFolder '_canonical_overlay']); 


ind = find(w<0.5);
reflectedPose = {images{ind}};
reflectedAsym = {Asym{nimage+ind}};
for ii=1:numel(reflectedAsym)
    reflectedAsym{ii} = flipdim(reflectedAsym{ii}, 2);
end

im = displayImagesAndSketches(reflectedPose, reflectedAsym, 5, sx, sy);
outepsgif(im, ['ex_flip_' posFolder '_reflected']); 

im = displayOverlay(reflectedPose, reflectedAsym, 8, sx, sy);

imwrite(im,'reflected_overlay.bmp');
outepsgif(im, ['ex_flip_' posFolder '_reflected_overlay']); 

towrite = learnedModel.sym;
towrite = uint8(255 * (towrite-min(towrite(:)))/(max(towrite(:))-min(towrite(:))));
imwrite(towrite,'template.bmp');
